/* 
   LZH8 compressor

   A straightforward implementation of LZSS, with output symbols Huffman coded.

   Note that the goal here is to reproduce exactly the compression used in
   Nintendo Virtual Console games. I've tried to point out with "POLICY:"
   comments places where I have determined particular behavioral oddities.
*/

#include <stdio.h>

#include "util.h"
#include "error_stuff.h"

#define VERSION "0.0 " __DATE__

#define SHOW_SYMBOLS        0
#define SHOW_FREQUENCIES    1
#define SHOW_TREE           1

#define LENCNT (1 << 9)
#define DISPCNT (1 << 5)

struct lzss_symbol
{
    uint8_t  is_reference;      /* flag, 0 = literal, 1 = reference */
    uint8_t length_or_literal; /* length of backreference -3, or literal byte */
    uint16_t offset;    /* displacement of reference -1, unused if literal */
};

struct huff_symbol
{
    int lchild, rchild;
    uint16_t key_len;
    uint16_t key_bits;
    uint16_t leaf;
};

void LZH8_compress(FILE *infile, FILE *outfile, long file_length);

void LZH8_LZSS_compress(
        unsigned char *input_data,
        long input_length,
        struct lzss_symbol **lzss_stream_p,
        long *lzss_length_p);

void LZH8_BuildHuff(
        struct lzss_symbol * const lzss_stream,
        long lzss_length,
        struct huff_symbol *back_litlen_table,
        struct huff_symbol *back_displen_table,
        long *output_offset,
        FILE *outfile);

void build_Huffman_tree(
    int *node_remains,
    long *freq,
    struct huff_symbol *node_array,
    int symbol_count);

void compute_prefix(
        struct huff_symbol *node_array,
        int root_idx,
        int key_bits,
        int key_len);

int LZH8_displen_length(uint16_t displacement);

int main(int argc, char **argv)
{
    if (argc != 3)
    {
        printf("lzh8_enc %s\n\n", VERSION);
        printf("Usage: %s infile outfile\n",argv[0]);
        exit(EXIT_FAILURE);
    }

    /* open file */
    FILE *infile = fopen(argv[1], "rb");
    CHECK_ERRNO(!infile, "fopen");

    FILE *outfile = fopen(argv[2], "wb");
    CHECK_ERRNO(!infile, "fopen");

    /* get file size */
    CHECK_ERRNO(fseek(infile, 0 , SEEK_END) != 0, "fseek");
    long file_length = ftell(infile);
    CHECK_ERRNO(file_length == -1, "ftell");

    rewind(infile);

    LZH8_compress(infile, outfile, file_length);

    CHECK_ERRNO(fclose(outfile) == EOF, "fclose");

    CHECK_ERRNO(fclose(infile) == EOF, "fclose");

    exit(EXIT_SUCCESS);
}

/* empty bit pool into outfile (if it wasn't empty already) */
static inline void flush_bits(
        FILE * const outfile,
        long * const offset_p,
        uint8_t * const bit_pool_p,
        int * const bits_written_p)
{
    if (0 != *bits_written_p)
    {
        put_byte_seek(*offset_p, *bit_pool_p, outfile);
        *bits_written_p = 0;
        *bit_pool_p = 0;
        ++*offset_p;
    }
}

/* write MSB->LSB order */
static inline void write_bits(
        FILE * const outfile,
        long * const offset_p,
        uint8_t * const bit_pool_p,
        int * const bits_written_p,
        const uint16_t bits_to_write,
        const int bit_count)
{
    int num_bits_produced = 0;
    while (num_bits_produced < bit_count)
    {
        if (8 == *bits_written_p)
        {
            flush_bits(outfile, offset_p, bit_pool_p, bits_written_p);
        }

        int bits_this_round;
        if (*bits_written_p + (bit_count - num_bits_produced) <= 8)
            bits_this_round = bit_count - num_bits_produced;
        else
            bits_this_round = 8 - *bits_written_p;

        *bit_pool_p |=
        (
        (bits_to_write >> (bit_count - bits_this_round - num_bits_produced)) & 
        ((1 << bits_this_round) - 1)
        ) << (8 - bits_this_round - num_bits_produced);

        *bits_written_p += bits_this_round;
        num_bits_produced += bits_this_round;
    }
}

#define GET_NEXT_BITS(bit_count) \
    get_next_bits(infile, &output_offset, &bit_pool, &bits_left, bit_count)

void LZH8_compress(FILE *infile, FILE *outfile, long file_length)
{
    long output_offset = 0;

    /*
       Step 0: Output header
    */
    {
        if (UINT32_C(0x1000000) > file_length)
        {
            put_32_le_seek( (((uint32_t)file_length) << 8) | 0x40,
                    output_offset, outfile );
            output_offset += 4;
        }
        else
        {
            /* >= 0x1000000 needs 4 extra bytes */

            CHECK_ERROR(file_length > UINT32_MAX, "input file is too large");

            put_32_le_seek( 0x40, output_offset, outfile );
            output_offset += 4;
            put_32_le_seek( file_length, output_offset, outfile );
            output_offset += 4;
        }
    }

    /*
       Step 1: LZSS with:
        reference length 3 <= length <= 2^8 - 1 + 3
        2^15 byte window
       Produce a series of symbols, uncompressed and compressed length+offset
    */
    struct lzss_symbol *lzss_stream = NULL;
    long lzss_length = 0;    /* length in symbols */

#if 0
    {
        unsigned char *input_data = NULL;

        input_data = malloc(file_length);
        CHECK_ERRNO( NULL == input_data, "malloc" );

        get_bytes_seek(0, infile, input_data, file_length);

        LZH8_LZSS_compress(input_data, file_length, &lzss_stream, &lzss_length);

#if 1
        // temp, dump LZSS stuff to file
        {
            FILE *lzss_dump = fopen("lzss.dump", "wb");
            CHECK_ERRNO(NULL == lzss_dump, "fopen");

            put_bytes(lzss_dump, (unsigned char *)lzss_stream,
                    lzss_length * sizeof(struct lzss_symbol));

            CHECK_ERRNO(EOF == fclose(lzss_dump), "fclose");
        }
#endif

        free(input_data);
    }
#else
    // temp, read LZSS stuff from file
    {
        FILE *lzss_dump = fopen("lzss.dump", "rb");
        CHECK_ERRNO(NULL == lzss_dump, "fopen");

        /* get dump size */
        CHECK_ERRNO(fseek(lzss_dump, 0 , SEEK_END) != 0, "fseek");
        long lzss_dump_length = ftell(lzss_dump);
        CHECK_ERRNO(lzss_dump_length == -1, "ftell");

        lzss_stream = malloc(lzss_dump_length);

        get_bytes_seek(0, lzss_dump,
                (unsigned char *)lzss_stream, lzss_dump_length);

        lzss_length = lzss_dump_length / sizeof(struct lzss_symbol);

        CHECK_ERRNO(EOF == fclose(lzss_dump), "fclose");
    }
#endif

    /*
       Step 2: Count frequencies and build Huffman codes, output flat trees
    */
    struct huff_symbol back_litlen_table[LENCNT*2];
    struct huff_symbol back_displen_table[DISPCNT*2];
    {
        LZH8_BuildHuff(lzss_stream, lzss_length, back_litlen_table,
                back_displen_table, &output_offset, outfile);
    }

    /*
       Step 3: Output the encoded symbol stream
    */
    {
    }

    free(lzss_stream);
}

void LZH8_LZSS_compress(
        unsigned char *input_data,
        long input_length,
        struct lzss_symbol **lzss_stream_p,
        long *lzss_length_p)
{
    /* parameters for this coding */
    const int min_length = 3;
    const int max_length = (1 << 8) - 1 + 3;
    const long max_window_size = (1l << 15);

    /* the output stream */
    struct lzss_symbol *lzss_stream = *lzss_stream_p;
    long lzss_length = *lzss_length_p;
    long lzss_stream_capacity = 0;
    CHECK_ERROR( NULL != lzss_stream || 0 != lzss_length,
            "should start with nothing");

    long bytes_done = 0;    /* bytes of input successfully encoded */
    long window_size = 0;   /* bytes (before bytes_done) of dictionary */
    long next_input_offset; /* next input byte to read */

    for (bytes_done = 0; bytes_done < input_length; )
    {
        int longest_match = 0;
        long longest_match_offset = 0;

        if (0 == lzss_length % (5*0x400))
        {
            fprintf(stderr,"%ld bytes done (%f%%)\n", bytes_done, (float)bytes_done/input_length*100);
        }
        
        /* grab as many input bytes as possible */
        next_input_offset = bytes_done + max_length;
        if (next_input_offset > input_length)
        {
            next_input_offset = input_length;
        }

        /* consider window */
        window_size = bytes_done;
        if (max_window_size < window_size)
        {
            window_size = max_window_size;
        }

        /* search for a match for what's currently in the input */
        for (long search_offset = bytes_done - 2,   /* POLICY: not -1 */
                  search_end = bytes_done - 1 - window_size;
             search_offset > search_end;
             search_offset--)
        {
            long match_length = 0;
            for (long match_check_offset = search_offset,
                      input_check_offset = bytes_done;
                 input_check_offset < next_input_offset &&
                 input_data[input_check_offset] ==
                    input_data[match_check_offset];
                 input_check_offset ++, match_check_offset++, match_length++)
            {}

            /* POLICY: prefer long matches, then first matches */
            if (match_length > longest_match)
            {
                longest_match = match_length;
                longest_match_offset = search_offset;
            }
        }

        /* check that there's room for a new symbol */
        if (lzss_length >= lzss_stream_capacity)
        {
            if (0 == lzss_stream_capacity)
                lzss_stream_capacity = 0x800;
            else
                lzss_stream_capacity *= 2;
            lzss_stream = realloc(lzss_stream,
                    lzss_stream_capacity*sizeof(struct lzss_symbol));
            CHECK_ERRNO( NULL == lzss_stream, "realloc" );
        }

        long bytes_in_this_symbol;

        /* record the new symbol */
        if (longest_match < min_length)
        {
            /* no backreference possible */
            lzss_stream[lzss_length].is_reference = 0;
            lzss_stream[lzss_length].length_or_literal = input_data[bytes_done];
            lzss_length++;
            bytes_in_this_symbol = 1;
        }
        else
        {
            /* generate a backreference */
            lzss_stream[lzss_length].is_reference = 1;
            lzss_stream[lzss_length].length_or_literal = longest_match - 3;
            lzss_stream[lzss_length].offset = bytes_done-longest_match_offset-1;
            lzss_length++;
            bytes_in_this_symbol = longest_match;
        }

        /* update state */
        bytes_done += bytes_in_this_symbol;
    }

#if SHOW_SYMBOLS
    for (long i = 0, offset = 0; i < lzss_length; i++)
    {
        printf("%08lx symbol %ld: ", (unsigned long)offset, i);
        if ( lzss_stream[i].is_reference )
        {
            printf("%d bytes, offset %d\n",
                    (int)lzss_stream[i].length_or_literal + 3,
                    (int)lzss_stream[i].offset+1);
            offset += lzss_stream[i].length_or_literal + 3;
        }
        else
        {
            printf("literal %02"PRIX8"\n", lzss_stream[i].length_or_literal);
            offset ++;
        }
    }
#endif

    *lzss_length_p = lzss_length;
    *lzss_stream_p = lzss_stream;
}

/* Count frequencies and build Huffman codes, output flat trees */
void LZH8_BuildHuff(
        struct lzss_symbol * const lzss_stream,
        long lzss_length,
        struct huff_symbol *back_litlen_table,
        struct huff_symbol *back_displen_table,
        long *output_offset,
        FILE *outfile)
{
    /* Count frequencies */
    long length_freq[LENCNT] = {0};
    long displen_freq[DISPCNT] = {0};

    for (long i=0; i < lzss_length; i++)
    {
        length_freq[ (lzss_stream[i].is_reference << 8) |
                      lzss_stream[i].length_or_literal ] ++;

        if (lzss_stream[i].is_reference)
        {
            displen_freq[ LZH8_displen_length(lzss_stream[i].offset) ] ++;
        }
    }

#if SHOW_FREQUENCIES
    for (int i=0; i < LENCNT; i++)
    {
        printf("%d: %ld\n", i, length_freq[i]);
    }
    for (int i=0; i < DISPCNT; i++)
    {
        printf("%d: %ld\n", i, displen_freq[i]);
    }
#endif

    /* Build Huffman codes for length/literal */
    {
#if SHOW_TREE
        printf("\nlength/literal tree:\n");
#endif
        int node_remains[LENCNT];
        build_Huffman_tree(
                node_remains, length_freq, back_litlen_table, LENCNT);
    }

    /* Build Huffman codes for displacement length */
    {
#if SHOW_TREE
        printf("\ndisplacement length tree:\n");
#endif
        int node_remains[DISPCNT];
        build_Huffman_tree(
                node_remains, displen_freq, back_displen_table, DISPCNT);
    }
}

void build_Huffman_tree(
    int *node_remains,
    long *freq,
    struct huff_symbol *node_array,
    int symbol_count
    )
{
    int nodes_left = 0;
    int new_node_idx = symbol_count;
    for (int i=0; i < symbol_count; i++)
    {
        if (0 != freq[i])
        {
            node_remains[i] = 1;
            nodes_left ++;
        }
        else
        {
            node_remains[i] = 0;
        }
        node_array[i].lchild = -1;
        node_array[i].rchild = -1;
        node_array[i].key_len = 0;
        node_array[i].key_bits = 0;
        node_array[i].leaf = i;
    }

    int root_idx;
    for (; nodes_left > 1; nodes_left --)
    {
        /* find smallest two (POLICY: favor infrequency, then low-index) */
        int smallest_idx, next_smallest_idx;
        {
            long smallest = -1, next_smallest = -1;
            for (int i=0; i < LENCNT; i++)
            {
                if (node_remains[i])
                {
                    if (freq[i] < smallest || -1 == smallest)
                    {
                        next_smallest = smallest;
                        next_smallest_idx = smallest_idx;
                        smallest = freq[i];
                        smallest_idx = i;
                    }
                    else if (freq[i] < next_smallest || -1 == next_smallest)
                    {
                        next_smallest = freq[i];
                        next_smallest_idx = i;
                    }
                }
            }
        }

        root_idx = next_smallest_idx;

        /* build new node to join the two smallest */
        freq[next_smallest_idx] += freq[smallest_idx];
        node_remains[smallest_idx] = 0;

        /* exchange nodes to put new leaf over at the end */
        struct huff_symbol temp = node_array[next_smallest_idx];
        node_array[next_smallest_idx] = node_array[new_node_idx];
        node_array[new_node_idx] = temp;

        /* set links */
        /* POLICY: smallest on left */
        node_array[next_smallest_idx].rchild = new_node_idx;
        node_array[next_smallest_idx].lchild = smallest_idx;

        new_node_idx ++;
    }

    compute_prefix(node_array, root_idx, 0, 0);
}

void compute_prefix(
        struct huff_symbol *node_array,
        int root_idx,
        int key_bits,
        int key_len)
{
    struct huff_symbol *root = &node_array[root_idx];
    CHECK_ERROR(
            (-1 == root->lchild && -1 != root->rchild) ||
            (-1 != root->lchild && -1 == root->rchild),
            "node not inner with two children or leaf");
    if (-1 != root->lchild)
    {
        key_len ++;
        compute_prefix(node_array, root->lchild, key_bits << 1, key_len);
        compute_prefix(node_array, root->rchild, (key_bits << 1) | 1, key_len);
    }
    else
    {
        root->key_len = key_len;
        root->key_bits = key_bits;
#if SHOW_TREE
        printf("%d: ", root->leaf);
        for (int i=key_len-1; i>=0; i--)
        {
            printf("%c", (key_bits&(1<<i)) ? '1' : '0');
        }
        printf("\n");
#endif
    }
}

int LZH8_displen_length(uint16_t displacement)
{
    int bits = 0;
    while (displacement)
    {
        displacement >>= 1;
        bits ++;
    }

    return bits;
}

